@@ -1,7 +1,7 @@
 /* options.c
 
 Copyright (C) 1999-2003 Tom Gilbert.
-Copyright (C) 2010-2011 Daniel Friesel.
+Copyright (C) 2010-2020 Daniel Friesel.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
@@ -24,6 +24,7 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 */
 
+#include <strings.h>
 #include "feh.h"
 #include "filelist.h"
 #include "options.h"
@@ -53,24 +54,31 @@ void init_parse_options(int argc, char **argv)
 	opt.display = 1;
 	opt.aspect = 1;
 	opt.slideshow_delay = 0.0;
-	opt.magick_timeout = -1;
+	opt.conversion_timeout = -1;
 	opt.thumb_w = 60;
 	opt.thumb_h = 60;
 	opt.thumb_redraw = 10;
 	opt.scroll_step = 20;
 	opt.menu_font = estrdup(DEFAULT_MENU_FONT);
 	opt.font = NULL;
-	opt.menu_bg = estrdup(PREFIX "/share/feh/images/menubg_default.png");
 	opt.max_height = opt.max_width = UINT_MAX;
 
+	opt.zoom_rate = 1.25;
+
 	opt.start_list_at = NULL;
 	opt.jump_on_resort = 1;
 
 	opt.screen_clip = 1;
+	opt.cache_size = 4;
 #ifdef HAVE_LIBXINERAMA
 	/* if we're using xinerama, then enable it by default */
 	opt.xinerama = 1;
+	opt.xinerama_index = -1;
 #endif				/* HAVE_LIBXINERAMA */
+#ifdef HAVE_INOTIFY
+	opt.auto_reload = 1;
+#endif				/* HAVE_INOTIFY */
+	opt.use_conversion_cache = 1;
 
 	feh_getopt_theme(argc, argv);
 
@@ -128,6 +136,7 @@ static void feh_load_options_for_theme(char *theme)
 	char *rcpath = NULL;
 	char *oldrcpath = NULL;
 	char *confbase = getenv("XDG_CONFIG_HOME");
+	// s, s1 and s2 must always have identical size
 	char s[1024], s1[1024], s2[1024];
 	int cont = 0;
 	int bspos;
@@ -164,11 +173,19 @@ static void feh_load_options_for_theme(char *theme)
 		s2[0] = '\0';
 
 		if (cont) {
+			/*
+			 * fgets ensures that s contains no more than 1023 characters
+			 * (+ 1 null byte)
+			 */
 			sscanf(s, " %[^\n]\n", (char *) &s2);
 			if (!*s2)
 				break;
 			D(("Got continued options %s\n", s2));
 		} else {
+			/*
+			 * fgets ensures that s contains no more than 1023 characters
+			 * (+ 1 null byte)
+			 */
 			sscanf(s, "%s %[^\n]\n", (char *) &s1, (char *) &s2);
 			if (!(*s1) || (!*s2) || (*s1 == '\n') || (*s1 == '#')) {
 				cont = 0;
@@ -210,7 +227,7 @@ static void feh_parse_options_from_string(char *opts)
 	char *s;
 	char *t;
 	char last = 0;
-	int inquote = 0;
+	char inquote = 0;
 	int i = 0;
 
 	/* So we don't reinvent the wheel (not again, anyway), we use the
@@ -225,7 +242,7 @@ static void feh_parse_options_from_string(char *opts)
 			eprintf(PACKAGE " does not support more than 64 words per "
 					"theme definition.\n Please shorten your lines.");
 
-		if ((*t == ' ') && !(inquote)) {
+		if ((*t == ' ') && !inquote) {
 			*t = '\0';
 			num++;
 
@@ -236,8 +253,10 @@ static void feh_parse_options_from_string(char *opts)
 
 			list[num - 1] = feh_string_normalize(s);
 			break;
-		} else if (((*t == '\"') || (*t == '\'')) && last != '\\')
-			inquote = !(inquote);
+		} else if ((*t == inquote) && (last != '\\')) {
+			inquote = 0;
+		} else if (((*t == '\"') || (*t == '\'')) && (last != '\\') && !inquote)
+			inquote = *t;
 		last = *t;
 	}
 
@@ -308,103 +327,116 @@ static void feh_parse_option_array(int argc, char **argv, int finalrun)
 {
 	int discard;
 	static char stropts[] =
-		"a:A:b:B:cC:dD:e:E:f:Fg:GhH:iIj:J:kK:lL:mM:nNo:O:pPqrR:sS:tT:uUvVwW:xXy:YzZ"
-		".@:^:~:):|:+:<:>:";
+		"a:A:b:B:C:dD:e:E:f:Fg:GhH:iIj:J:kK:lL:mM:nNo:O:pPqrR:sS:tT:uUvVwW:xXy:YzZ"
+		".@:^:~:|:+:<:>:";
 
 	/* (*name, has_arg, *flag, val) See: struct option in getopts.h */
 	static struct option lopts[] = {
-		{"menu-bg"       , 1, 0, ')'},
-		{"debug"         , 0, 0, '+'},
-		{"scale-down"    , 0, 0, '.'},
-		{"max-dimension" , 1, 0, '<'},
-		{"min-dimension" , 1, 0, '>'},
-		{"title-font"    , 1, 0, '@'},
-		{"action"        , 1, 0, 'A'},
-		{"image-bg"      , 1, 0, 'B'},
-		{"fontpath"      , 1, 0, 'C'},
-		{"slideshow-delay",1, 0, 'D'},
-		{"thumb-height"  , 1, 0, 'E'},
-		{"full-screen"   , 0, 0, 'F'}, /* deprecated */
-		{"fullscreen"    , 0, 0, 'F'},
-		{"draw-actions"  , 0, 0, 'G'},
-		{"limit-height"  , 1, 0, 'H'},
-		{"fullindex"     , 0, 0, 'I'},
-		{"thumb-redraw"  , 1, 0, 'J'},
-		{"caption-path"  , 1, 0, 'K'},
-		{"customlist"    , 1, 0, 'L'},
-		{"menu-font"     , 1, 0, 'M'},
-		{"no-menus"      , 0, 0, 'N'},
-		{"output-only"   , 1, 0, 'O'},
-		{"cache-thumbnails", 0, 0, 'P'},
-		{"reload"        , 1, 0, 'R'},
-		{"sort"          , 1, 0, 'S'},
-		{"theme"         , 1, 0, 'T'},
-		{"loadable"      , 0, 0, 'U'},
-		{"verbose"       , 0, 0, 'V'},
-		{"limit-width"   , 1, 0, 'W'},
-		{"ignore-aspect" , 0, 0, 'X'},
-		{"hide-pointer"  , 0, 0, 'Y'},
-		{"auto-zoom"     , 0, 0, 'Z'},
-		{"title"         , 1, 0, '^'},
-		{"alpha"         , 1, 0, 'a'},
-		{"bg"            , 1, 0, 'b'},
-		{"collage"       , 0, 0, 'c'},
-		{"draw-filename" , 0, 0, 'd'},
-		{"font"          , 1, 0, 'e'},
-		{"filelist"      , 1, 0, 'f'},
-		{"geometry"      , 1, 0, 'g'},
-		{"help"          , 0, 0, 'h'},
-		{"index"         , 0, 0, 'i'},
-		{"output-dir"    , 1, 0, 'j'},
-		{"keep-http"     , 0, 0, 'k'},
-		{"list"          , 0, 0, 'l'},
-		{"montage"       , 0, 0, 'm'},
-		{"reverse"       , 0, 0, 'n'},
-		{"output"        , 1, 0, 'o'},
-		{"preload"       , 0, 0, 'p'},
-		{"quiet"         , 0, 0, 'q'},
-		{"recursive"     , 0, 0, 'r'},
-		{"stretch"       , 0, 0, 's'},
-		{"thumbnails"    , 0, 0, 't'},
-		{"unloadable"    , 0, 0, 'u'},
-		{"version"       , 0, 0, 'v'},
-		{"multiwindow"   , 0, 0, 'w'},
-		{"borderless"    , 0, 0, 'x'},
-		{"thumb-width"   , 1, 0, 'y'},
-		{"randomize"     , 0, 0, 'z'},
-		{"start-at"      , 1, 0, '|'},
-		{"thumb-title"   , 1, 0, '~'},
-		{"bg-tile"       , 0, 0, 200},
-		{"bg-center"     , 0, 0, 201},
-		{"bg-scale"      , 0, 0, 202},
-		{"zoom"          , 1, 0, 205},
-		{"no-screen-clip", 0, 0, 206},
-		{"index-info"    , 1, 0, 207},
-		{"magick-timeout", 1, 0, 208},
-		{"action1"       , 1, 0, 209},
-		{"action2"       , 1, 0, 210},
-		{"action3"       , 1, 0, 211},
-		{"action4"       , 1, 0, 212},
-		{"action5"       , 1, 0, 213},
-		{"action6"       , 1, 0, 214},
-		{"action7"       , 1, 0, 215},
-		{"action8"       , 1, 0, 216},
-		{"action9"       , 1, 0, 217},
-		{"bg-fill"       , 0, 0, 218},
-		{"bg-max"        , 0, 0, 219},
-		{"no-jump-on-resort", 0, 0, 220},
+		{"debug"         , 0, 0, OPTION_debug},
+		{"scale-down"    , 0, 0, OPTION_scale_down},
+		{"max-dimension" , 1, 0, OPTION_max_dimension},
+		{"min-dimension" , 1, 0, OPTION_min_dimension},
+		{"title-font"    , 1, 0, OPTION_title_font},
+		{"action"        , 1, 0, OPTION_action},
+		{"image-bg"      , 1, 0, OPTION_image_bg},
+		{"fontpath"      , 1, 0, OPTION_fontpath},
+		{"slideshow-delay",1, 0, OPTION_slideshow_delay},
+		{"thumb-height"  , 1, 0, OPTION_thumb_height},
+		{"full-screen"   , 0, 0, OPTION_fullscreen}, /* deprecated */
+		{"fullscreen"    , 0, 0, OPTION_fullscreen},
+		{"draw-actions"  , 0, 0, OPTION_draw_actions},
+		{"limit-height"  , 1, 0, OPTION_limit_height},
+		{"fullindex"     , 0, 0, OPTION_fullindex},
+		{"thumb-redraw"  , 1, 0, OPTION_thumb_redraw},
+		{"caption-path"  , 1, 0, OPTION_caption_path},
+		{"customlist"    , 1, 0, OPTION_customlist},
+		{"menu-font"     , 1, 0, OPTION_menu_font},
+		{"no-menus"      , 0, 0, OPTION_no_menus},
+		{"output-only"   , 1, 0, OPTION_output_only},
+		{"cache-thumbnails", 0, 0, OPTION_cache_thumbnails},
+		{"reload"        , 1, 0, OPTION_reload},
+		{"sort"          , 1, 0, OPTION_sort},
+		{"theme"         , 1, 0, OPTION_theme},
+		{"loadable"      , 0, 0, OPTION_loadable},
+		{"verbose"       , 0, 0, OPTION_verbose},
+		{"limit-width"   , 1, 0, OPTION_limit_width},
+		{"ignore-aspect" , 0, 0, OPTION_ignore_aspect},
+		{"hide-pointer"  , 0, 0, OPTION_hide_pointer},
+		{"auto-zoom"     , 0, 0, OPTION_auto_zoom},
+		{"title"         , 1, 0, OPTION_title},
+		{"alpha"         , 1, 0, OPTION_alpha},
+		{"bg"            , 1, 0, OPTION_bg},
+		{"draw-filename" , 0, 0, OPTION_draw_filename},
+		{"font"          , 1, 0, OPTION_font},
+		{"filelist"      , 1, 0, OPTION_filelist},
+		{"geometry"      , 1, 0, OPTION_geometry},
+		{"help"          , 0, 0, OPTION_help},
+		{"index"         , 0, 0, OPTION_index},
+		{"output-dir"    , 1, 0, OPTION_output_dir},
+		{"keep-http"     , 0, 0, OPTION_keep_http},
+		{"list"          , 0, 0, OPTION_list},
+		{"montage"       , 0, 0, OPTION_montage},
+		{"reverse"       , 0, 0, OPTION_reverse},
+		{"output"        , 1, 0, OPTION_output},
+		{"preload"       , 0, 0, OPTION_preload},
+		{"quiet"         , 0, 0, OPTION_quiet},
+		{"recursive"     , 0, 0, OPTION_recursive},
+		{"stretch"       , 0, 0, OPTION_stretch},
+		{"thumbnails"    , 0, 0, OPTION_thumbnails},
+		{"unloadable"    , 0, 0, OPTION_unloadable},
+		{"version"       , 0, 0, OPTION_version},
+		{"multiwindow"   , 0, 0, OPTION_multiwindow},
+		{"borderless"    , 0, 0, OPTION_borderless},
+		{"thumb-width"   , 1, 0, OPTION_thumb_width},
+		{"randomize"     , 0, 0, OPTION_randomize},
+		{"start-at"      , 1, 0, OPTION_start_at},
+		{"thumb-title"   , 1, 0, OPTION_thumb_title},
+		{"bg-tile"       , 0, 0, OPTION_bg_title},
+		{"bg-center"     , 0, 0, OPTION_bg_center},
+		{"bg-scale"      , 0, 0, OPTION_bg_scale},
+		{"zoom"          , 1, 0, OPTION_zoom},
+		{"zoom-step"     , 1, 0, OPTION_zoom_step},
+		{"no-screen-clip", 0, 0, OPTION_no_screen_clip},
+		{"index-info"    , 1, 0, OPTION_index_info},
+		{"magick-timeout", 1, 0, OPTION_magick_timeout},
+		{"action1"       , 1, 0, OPTION_action1},
+		{"action2"       , 1, 0, OPTION_action2},
+		{"action3"       , 1, 0, OPTION_action3},
+		{"action4"       , 1, 0, OPTION_action4},
+		{"action5"       , 1, 0, OPTION_action5},
+		{"action6"       , 1, 0, OPTION_action6},
+		{"action7"       , 1, 0, OPTION_action7},
+		{"action8"       , 1, 0, OPTION_action8},
+		{"action9"       , 1, 0, OPTION_action9},
+		{"bg-fill"       , 0, 0, OPTION_bg_fill},
+		{"bg-max"        , 0, 0, OPTION_bg_max},
+		{"no-jump-on-resort", 0, 0, OPTION_no_jump_on_resort},
+		{"edit"          , 0, 0, OPTION_edit},
 #ifdef HAVE_LIBEXIF
-		{"draw-exif"     , 0, 0, 223},
+		{"draw-exif"     , 0, 0, OPTION_draw_exif},
+		{"auto-rotate"   , 0, 0, OPTION_auto_rotate},
 #endif
-		{"cycle-once"    , 0, 0, 224},
-		{"no-xinerama"   , 0, 0, 225},
-		{"draw-tinted"   , 0, 0, 229},
-		{"info"          , 1, 0, 234},
-		{"force-aliasing", 0, 0, 235},
-		{"no-fehbg"      , 0, 0, 236},
-		{"keep-zoom-vp"  , 0, 0, 237},
-		{"scroll-step"   , 1, 0, 238},
-
+		{"no-xinerama"   , 0, 0, OPTION_no_xinerama},
+		{"draw-tinted"   , 0, 0, OPTION_draw_tinted},
+		{"info"          , 1, 0, OPTION_info},
+		{"force-aliasing", 0, 0, OPTION_force_aliasing},
+		{"no-fehbg"      , 0, 0, OPTION_no_fehbg},
+		{"keep-zoom-vp"  , 0, 0, OPTION_keep_zoom_vp},
+		{"scroll-step"   , 1, 0, OPTION_scroll_step},
+		{"xinerama-index", 1, 0, OPTION_xinerama_index},
+		{"insecure"      , 0, 0, OPTION_insecure},
+		{"no-recursive"  , 0, 0, OPTION_recursive},
+		{"cache-size"    , 1, 0, OPTION_cache_size},
+		{"on-last-slide" , 1, 0, OPTION_on_last_slide},
+		{"conversion-timeout" , 1, 0, OPTION_conversion_timeout},
+		{"version-sort"  , 0, 0, OPTION_version_sort},
+		{"offset"        , 1, 0, OPTION_offset},
+#ifdef HAVE_INOTIFY
+		{"auto-reload"   , 0, 0, OPTION_auto_reload},
+#endif
+		{"class"         , 1, 0, OPTION_class},
+		{"no-conversion-cache", 0, 0, OPTION_no_conversion_cache},
+		{"window-id", 1, 0, OPTION_window_id},
 		{0, 0, 0, 0}
 	};
 	int optch = 0, cmdx = 0;
@@ -414,103 +446,101 @@ static void feh_parse_option_array(int argc, char **argv, int finalrun)
 		switch (optch) {
 		case 0:
 			break;
-		case ')':
-			free(opt.menu_bg);
-			opt.menu_bg = estrdup(optarg);
-			weprintf("The --menu-bg option is deprecated and will be removed by 2012");
-			break;
-		case '+':
+		case OPTION_debug:
 			opt.debug = 1;
 			break;
-		case '<':
+		case OPTION_max_dimension:
+			opt.filter_by_dimensions = 1;
 			XParseGeometry(optarg, &discard, &discard, &opt.max_width, &opt.max_height);
 			if (opt.max_width == 0)
 				opt.max_width = UINT_MAX;
 			if (opt.max_height == 0)
 				opt.max_height = UINT_MAX;
 			break;
-		case '>':
+		case OPTION_min_dimension:
+			opt.filter_by_dimensions = 1;
 			XParseGeometry(optarg, &discard, &discard, &opt.min_width, &opt.min_height);
 			break;
-		case '.':
+		case OPTION_scale_down:
 			opt.scale_down = 1;
 			break;
-		case '@':
+		case OPTION_title_font:
 			opt.title_font = estrdup(optarg);
 			break;
-		case 'A':
+		case OPTION_action:
 			opt.actions[0] = estrdup(optarg);
 			break;
-		case 'B':
-			if (!strcmp(optarg, "checks"))
-				opt.image_bg = IMAGE_BG_CHECKS;
-			else if (!strcmp(optarg, "white"))
-				opt.image_bg = IMAGE_BG_WHITE;
-			else if (!strcmp(optarg, "black"))
-				opt.image_bg = IMAGE_BG_BLACK;
-			else
-				weprintf("Unknown argument to --image-bg: %s", optarg);
+		case OPTION_image_bg:
+			opt.image_bg = estrdup(optarg);
 			break;
-		case 'C':
+		case OPTION_fontpath:
 			D(("adding fontpath %s\n", optarg));
 			imlib_add_path_to_font_path(optarg);
 			break;
-		case 'D':
+		case OPTION_slideshow_delay:
 			opt.slideshow_delay = atof(optarg);
 			if (opt.slideshow_delay < 0.0) {
 				opt.slideshow_delay *= (-1);
 				opt.paused = 1;
+			} else {
+				opt.paused = 0;
 			}
 			break;
-		case 'E':
+		case OPTION_thumb_height:
 			opt.thumb_h = atoi(optarg);
 			break;
-		case 'F':
+		case OPTION_fullscreen:
 			opt.full_screen = 1;
 			break;
-		case 'G':
+		case OPTION_draw_actions:
 			opt.draw_actions = 1;
 			break;
-		case 'H':
+		case OPTION_limit_height:
 			opt.limit_h = atoi(optarg);
 			break;
-		case 'I':
+		case OPTION_fullindex:
 			opt.index = 1;
 			opt.index_info = estrdup("%n\n%S\n%wx%h");
 			break;
-		case 'J':
+		case OPTION_thumb_redraw:
 			opt.thumb_redraw = atoi(optarg);
 			break;
-		case 'K':
+		case OPTION_caption_path:
 			opt.caption_path = estrdup(optarg);
 			break;
-		case 'L':
+		case OPTION_customlist:
 			opt.customlist = estrdup(optarg);
 			opt.display = 0;
 			break;
-		case 'M':
+		case OPTION_menu_font:
 			free(opt.menu_font);
 			opt.menu_font = estrdup(optarg);
 			break;
-		case 'N':
+		case OPTION_no_menus:
 			opt.no_menus = 1;
 			break;
-		case 'O':
+		case OPTION_output_only:
 			opt.output = 1;
 			opt.output_file = estrdup(optarg);
 			opt.display = 0;
 			break;
-		case 'P':
+		case OPTION_cache_thumbnails:
 			opt.cache_thumbnails = 1;
 			break;
-		case 'R':
+		case OPTION_reload:
 			opt.reload = atof(optarg);
+			opt.use_conversion_cache = 0;
+#ifdef HAVE_INOTIFY
+			opt.auto_reload = 0;
+#endif
 			break;
-		case 'S':
+		case OPTION_sort:
 			if (!strcasecmp(optarg, "name"))
 				opt.sort = SORT_NAME;
 			else if (!strcasecmp(optarg, "filename"))
 				opt.sort = SORT_FILENAME;
+			else if (!strcasecmp(optarg, "dirname"))
+				opt.sort = SORT_DIRNAME;
 			else if (!strcasecmp(optarg, "mtime"))
 				opt.sort = SORT_MTIME;
 			else if (!strcasecmp(optarg, "width"))
@@ -528,137 +558,145 @@ static void feh_parse_option_array(int argc, char **argv, int finalrun)
 						"sort by filename", optarg);
 				opt.sort = SORT_FILENAME;
 			}
+			if (opt.randomize) {
+				weprintf("commandline contains --randomize and --sort. "
+						"--randomize has been unset");
+				opt.randomize = 0;
+			}
 			break;
-		case 'T':
+		case OPTION_theme:
 			theme = estrdup(optarg);
 			break;
-		case 'U':
+		case OPTION_loadable:
 			opt.loadables = 1;
 			opt.display = 0;
 			break;
-		case 'V':
+		case OPTION_verbose:
 			opt.verbose = 1;
 			break;
-		case 'W':
+		case OPTION_limit_width:
 			opt.limit_w = atoi(optarg);
 			break;
-		case 'X':
+		case OPTION_ignore_aspect:
 			opt.aspect = 0;
 			break;
-		case 'Y':
+		case OPTION_hide_pointer:
 			opt.hide_pointer = 1;
 			break;
-		case 'Z':
+		case OPTION_auto_zoom:
 			opt.zoom_mode = ZOOM_MODE_MAX;
 			break;
-		case '^':
+		case OPTION_title:
 			opt.title = estrdup(optarg);
 			break;
-		case 'a':
+		case OPTION_alpha:
 			opt.alpha = 1;
 			opt.alpha_level = 255 - atoi(optarg);
 			break;
-		case 'b':
+		case OPTION_bg:
 			opt.bg = 1;
 			opt.bg_file = estrdup(optarg);
 			break;
-		case 'c':
-			opt.collage = 1;
-			break;
-		case 'd':
+		case OPTION_draw_filename:
 			opt.draw_filename = 1;
 			break;
-		case 'e':
+		case OPTION_font:
 			opt.font = estrdup(optarg);
 			break;
-		case 'f':
+		case OPTION_filelist:
 			if (!strcmp(optarg, "-"))
 				opt.filelistfile = estrdup("/dev/stdin");
 			else
 				opt.filelistfile = estrdup(optarg);
 			break;
-		case 'g':
+		case OPTION_geometry:
+			opt.geom_enabled = 1;
 			opt.geom_flags = XParseGeometry(optarg, &opt.geom_x,
 					&opt.geom_y, &opt.geom_w, &opt.geom_h);
 			break;
-		case 'h':
+		case OPTION_help:
 			show_usage();
 			break;
-		case 'i':
+		case OPTION_index:
 			opt.index = 1;
 			opt.index_info = estrdup("%n");
 			break;
-		case 'j':
+		case OPTION_output_dir:
 			opt.output_dir = estrdup(optarg);
 			break;
-		case 'k':
+		case OPTION_keep_http:
 			opt.keep_http = 1;
 			break;
-		case 'l':
+		case OPTION_list:
 			opt.list = 1;
 			opt.display = 0;
 			break;
-		case 'm':
+		case OPTION_montage:
 			opt.index = 1;
 			break;
-		case 'n':
+		case OPTION_reverse:
 			opt.reverse = 1;
 			break;
-		case 'o':
+		case OPTION_output:
 			opt.output = 1;
 			opt.output_file = estrdup(optarg);
 			break;
-		case 'p':
+		case OPTION_preload:
 			opt.preload = 1;
 			break;
-		case 'q':
+		case OPTION_quiet:
 			opt.quiet = 1;
 			break;
-		case 'r':
+		case OPTION_recursive:
 			opt.recursive = 1;
 			break;
-		case 's':
+		case OPTION_stretch:
 			opt.stretch = 1;
 			break;
-		case 't':
+		case OPTION_thumbnails:
 			opt.thumbs = 1;
 			opt.index_info = estrdup("%n");
 			break;
-		case 'u':
+		case OPTION_unloadable:
 			opt.unloadables = 1;
 			opt.display = 0;
 			break;
-		case 'v':
+		case OPTION_version:
 			show_version();
 			break;
-		case 'w':
+		case OPTION_multiwindow:
 			opt.multiwindow = 1;
 			break;
-		case 'x':
+		case OPTION_borderless:
 			opt.borderless = 1;
 			break;
-		case 'y':
+		case OPTION_thumb_width:
 			opt.thumb_w = atoi(optarg);
 			break;
-		case 'z':
+		case OPTION_randomize:
 			opt.randomize = 1;
+			if (opt.sort != SORT_NONE) {
+				weprintf("commandline contains --sort and --randomize. "
+						"--sort has been unset");
+				opt.sort = SORT_NONE;
+			}
 			break;
-		case '|':
+		case OPTION_start_at:
 			opt.start_list_at = estrdup(optarg);
 			break;
-		case '~':
+		case OPTION_thumb_title:
 			opt.thumb_title = estrdup(optarg);
 			break;
-		case 200:
+		case OPTION_bg_title:
 			opt.bgmode = BG_MODE_TILE;
 			break;
-		case 201:
+		case OPTION_bg_center:
 			opt.bgmode = BG_MODE_CENTER;
 			break;
-		case 202:
+		case OPTION_bg_scale:
 			opt.bgmode = BG_MODE_SCALE;
 			break;
-		case 205:
+		case OPTION_zoom:
 			if (!strcmp("fill", optarg))
 				opt.zoom_mode = ZOOM_MODE_FILL;
 			else if (!strcmp("max", optarg))
@@ -666,84 +704,151 @@ static void feh_parse_option_array(int argc, char **argv, int finalrun)
 			else
 				opt.default_zoom = atoi(optarg);
 			break;
-		case 206:
+		case OPTION_no_screen_clip:
 			opt.screen_clip = 0;
 			break;
-		case 207:
+		case OPTION_index_info:
 			opt.index_info = estrdup(optarg);
 			break;
-		case 208:
-			opt.magick_timeout = atoi(optarg);
+		case OPTION_magick_timeout:
+			weprintf("--magick-timeout is deprecated, please use --conversion-timeout instead");
+			opt.conversion_timeout = atoi(optarg);
 			break;
-		case 209:
+		case OPTION_action1:
 			opt.actions[1] = estrdup(optarg);
 			break;
-		case 210:
+		case OPTION_action2:
 			opt.actions[2] = estrdup(optarg);
 			break;
-		case 211:
+		case OPTION_action3:
 			opt.actions[3] = estrdup(optarg);
 			break;
-		case 212:
+		case OPTION_action4:
 			opt.actions[4] = estrdup(optarg);
 			break;
-		case 213:
+		case OPTION_action5:
 			opt.actions[5] = estrdup(optarg);
 			break;
-		case 214:
+		case OPTION_action6:
 			opt.actions[6] = estrdup(optarg);
 			break;
-		case 215:
+		case OPTION_action7:
 			opt.actions[7] = estrdup(optarg);
 			break;
-		case 216:
+		case OPTION_action8:
 			opt.actions[8] = estrdup(optarg);
 			break;
-		case 217:
+		case OPTION_action9:
 			opt.actions[9] = estrdup(optarg);
 			break;
-		case 218:
+		case OPTION_bg_fill:
 			opt.bgmode = BG_MODE_FILL;
 			break;
-		case 219:
+		case OPTION_bg_max:
 			opt.bgmode = BG_MODE_MAX;
 			break;
-		case 220:
+		case OPTION_no_jump_on_resort:
 			opt.jump_on_resort = 0;
 			break;
+		case OPTION_edit:
+			opt.edit = 1;
+			break;
 #ifdef HAVE_LIBEXIF
-		case 223:
+		case OPTION_draw_exif:
 			opt.draw_exif = 1;
 			break;
-#endif
-		case 224:
-			opt.cycle_once = 1;
+		case OPTION_auto_rotate:
+			opt.auto_rotate = 1;
 			break;
-		case 225:
+#endif
+		case OPTION_no_xinerama:
 			opt.xinerama = 0;
 			break;
-		case 229:
+		case OPTION_draw_tinted:
 			opt.text_bg = TEXT_BG_TINTED;
 			break;
-		case 234:
+		case OPTION_info:
 			opt.info_cmd = estrdup(optarg);
-			if (opt.info_cmd[0] == ';')
+			if (opt.info_cmd[0] == ';') {
+				opt.draw_info = 0;
 				opt.info_cmd++;
-			else
+			} else {
 				opt.draw_info = 1;
+			}
 			break;
-		case 235:
+		case OPTION_force_aliasing:
 			opt.force_aliasing = 1;
 			break;
-		case 236:
+		case OPTION_no_fehbg:
 			opt.no_fehbg = 1;
 			break;
-		case 237:
+		case OPTION_keep_zoom_vp:
 			opt.keep_zoom_vp = 1;
 			break;
-		case 238:
+		case OPTION_scroll_step:
 			opt.scroll_step = atoi(optarg);
 			break;
+		case OPTION_xinerama_index:
+			opt.xinerama_index = atoi(optarg);
+			break;
+		case OPTION_insecure:
+			opt.insecure_ssl = 1;
+			break;
+		case OPTION_no_recursive:
+			opt.recursive = 0;
+			break;
+		case OPTION_cache_size:
+			opt.cache_size = atoi(optarg);
+			if (opt.cache_size < 0)
+				opt.cache_size = 0;
+			if (opt.cache_size > 2048)
+				opt.cache_size = 2048;
+			break;
+		case OPTION_on_last_slide:
+			if (!strcmp(optarg, "quit")) {
+				opt.on_last_slide = ON_LAST_SLIDE_QUIT;
+			} else if (!strcmp(optarg, "hold")) {
+				opt.on_last_slide = ON_LAST_SLIDE_HOLD;
+			} else if (!strcmp(optarg, "resume")) {
+				opt.on_last_slide = ON_LAST_SLIDE_RESUME;
+			} else {
+				weprintf("Unrecognized on-last-slide action \"%s\"."
+						"Supported actions: hold, resume, quit\n", optarg);
+			}
+			break;
+		case OPTION_conversion_timeout:
+			opt.conversion_timeout = atoi(optarg);
+			break;
+		case OPTION_version_sort:
+			opt.version_sort = 1;
+			break;
+		case OPTION_offset:
+			opt.offset_flags = XParseGeometry(optarg, &opt.offset_x,
+					&opt.offset_y, (unsigned int *)&discard, (unsigned int *)&discard);
+			break;
+#ifdef HAVE_INOTIFY
+		case OPTION_auto_reload:
+			opt.auto_reload = 1;
+			break;
+#endif
+		case OPTION_class:
+			opt.x11_class = estrdup(optarg);
+			break;
+		case OPTION_no_conversion_cache:
+			opt.use_conversion_cache = 0;
+			break;
+		case OPTION_window_id:
+			opt.x11_windowid = atol(optarg);
+			break;
+		case OPTION_zoom_step:
+			opt.zoom_rate = atof(optarg);
+			if ((opt.zoom_rate <= 0)) {
+				weprintf("Zooming disabled due to --zoom-step=%f", opt.zoom_rate);
+				opt.zoom_rate = 1.0;
+			} else {
+				opt.zoom_rate = 1 + ((float)opt.zoom_rate / 100);
+			}
+			break;
 		default:
 			break;
 		}
@@ -759,8 +864,46 @@ static void feh_parse_option_array(int argc, char **argv, int finalrun)
 			add_file_to_filelist_recursively(argv[optind++], FILELIST_FIRST);
 		}
 	}
-	else if (finalrun && !opt.filelistfile && !opt.bgmode)
-		add_file_to_filelist_recursively(".", FILELIST_FIRST);
+	else if (finalrun && !opt.filelistfile && !opt.bgmode) {
+		/*
+		 * if --start-at is a non-local URL (i.e., does not start with file:///),
+		 * behave as if "feh URL" was called (there is no directory we can load)
+		 */
+		if (opt.start_list_at && path_is_url(opt.start_list_at) && (strlen(opt.start_list_at) <= 8 || strncmp(opt.start_list_at, "file:///", 8) != 0)) {
+			add_file_to_filelist_recursively(opt.start_list_at, FILELIST_FIRST);
+		/*
+		 * Otherwise, make "feh --start-at dir/file.jpg" behave like
+		 * "feh --start-at dir/file.jpg dir".
+		 */
+		} else if (opt.start_list_at && strrchr(opt.start_list_at, '/')) {
+			/*
+			 * feh can't candle urlencoded path components ("some%20dir" etc).
+			 * Use libcurl to unescape them if --start-at is file://...
+			 */
+			if (strlen(opt.start_list_at) > 8 && strncmp(opt.start_list_at, "file:///", 8) == 0) {
+				char *unescaped_path = feh_http_unescape(opt.start_list_at);
+				if (unescaped_path != NULL) {
+					free(opt.start_list_at);
+					opt.start_list_at = estrdup(unescaped_path + 7);
+					free(unescaped_path);
+				} else {
+					char *new_path = estrdup(opt.start_list_at + 7);
+					free(opt.start_list_at);
+					opt.start_list_at = new_path;
+				}
+			}
+			char *target_directory = estrdup(opt.start_list_at);
+			char *filename_start = strrchr(target_directory, '/');
+			if (filename_start) {
+				*filename_start = '\0';
+			}
+			add_file_to_filelist_recursively(target_directory, FILELIST_FIRST);
+			original_file_items = gib_list_add_front(original_file_items, estrdup(target_directory));
+			free(target_directory);
+		} else {
+			add_file_to_filelist_recursively(".", FILELIST_FIRST);
+		}
+	}
 
 	/* So that we can safely be called again */
 	optind = 0;
@@ -770,33 +913,32 @@ static void feh_parse_option_array(int argc, char **argv, int finalrun)
 static void check_options(void)
 {
 	int i;
+	char *endptr;
+
 	for (i = 0; i < 10; i++) {
 		if (opt.actions[i] && !opt.hold_actions[i] && (opt.actions[i][0] == ';')) {
 			opt.hold_actions[i] = 1;
-			opt.actions[i] = &opt.actions[i][1];
+			opt.actions[i] = opt.actions[i] + 1;
+		}
+		opt.action_titles[i] = opt.actions[i];
+		if (opt.actions[i] && (opt.actions[i][0] == '[')) {
+			if (((endptr = strchr(opt.actions[i], ']')) != NULL)
+					&& (opt.actions[i][1] != ' ')) {
+				opt.action_titles[i] = opt.actions[i] + 1;
+				opt.actions[i] = endptr + 1;
+				*endptr = 0;
+			}
 		}
-	}
-
-	if ((opt.index + opt.collage) > 1) {
-		weprintf("you can't use collage mode and index mode together.\n"
-				"   I'm going with index");
-		opt.collage = 0;
 	}
 
 	if (opt.full_screen && opt.multiwindow) {
 		eprintf("You cannot combine --fullscreen with --multiwindow");
 	}
 
-	if (opt.list && (opt.multiwindow || opt.index || opt.collage)) {
+	if (opt.list && (opt.multiwindow || opt.index)) {
 		eprintf("You cannot combine --list with other modes");
 	}
 
-	if (opt.sort && opt.randomize) {
-		weprintf("You cant sort AND randomize the filelist...\n"
-				"randomize mode has been unset\n");
-		opt.randomize = 0;
-	}
-
 	if (opt.loadables && opt.unloadables) {
 		eprintf("You cannot combine --loadable with --unloadable");
 	}
@@ -821,6 +963,10 @@ static void show_version(void)
 		"exif "
 #endif
 
+#ifdef HAVE_INOTIFY
+		"inotify "
+#endif
+
 #ifdef INCLUDE_HELP
 		"help "
 #endif
@@ -829,6 +975,10 @@ static void show_version(void)
 		"stat64 "
 #endif
 
+#ifdef HAVE_STRVERSCMP
+         "verscmp "
+#endif
+
 #ifdef HAVE_LIBXINERAMA
 		"xinerama "
 #endif
